(******************************************************************************)
(* computer (stack machine)                                                   *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;
FROM Comp0 IMPORT TValue, TIndex, TOperation, TFunction;
FROM Comp0 IMPORT TCmdDesc, TSet, TCmdStr;
FROM Comp0 IMPORT acMax, vcMax, pcMax, vFalse, vTrue, aNone;
FROM Comp0 IMPORT Convert;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

VAR
  (* program *)
  prg: ARRAY [ 0 .. pcMax - 1 ] OF BYTE;
  pos, len: TAddress;
  (* machine state (execution) *)
  cmd: TCmdDesc;
  pc, pc0: CARDINAL;
  args: ARRAY [ 1 .. acMax ] OF TValue;  (* stack *)
  ac: CARDINAL;
  vars: ARRAY [ 0 .. vcMax - 1 ] OF TValue;  (* data *)
  (* module state *)
  strs: ARRAY BOOLEAN OF ARRAY [ 0 .. 4 ] OF CHAR;
  vals: ARRAY BOOLEAN OF TValue;
  first: BOOLEAN;

(***** debugging *****)

PROCEDURE Dbg();
VAR s: TCmdStr;
BEGIN
  Convert( cmd, s );
  IF first THEN
    FIO.WrStr( dbg, 'Comp.Dbg:   pc  ac       top-1         top' );
    FIO.WrLn( dbg );
    first := FALSE;
  END;
  FIO.WrStr( dbg, 'Comp.Dbg: ' );
  FIO.WrCard( dbg, pc0, 4 );
  FIO.WrStr( dbg, ': ' );
  FIO.WrCard( dbg, ac, 2 );
  FIO.WrChar( dbg, ' ' );
  IF ac = 0 THEN
    FIO.WrStr( dbg, '                       ' );
  ELSIF ac = 1 THEN
    FIO.WrStr( dbg, '            ' );
    FIO.WrLngInt( dbg, args[ac], 11 );
  ELSE
    FIO.WrLngInt( dbg, args[ac-1], 11 );
    FIO.WrChar( dbg, ' ' );
    FIO.WrLngInt( dbg, args[ac], 11 );
  END;
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END Dbg;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ErrCard( s: ARRAY OF CHAR; x: CARDINAL );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrChar( err, ' ' );
  FIO.WrCard( err, x, 0 );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END ErrCard;

(***** execution *****)

PROCEDURE Fetch( VAR x: ARRAY OF BYTE ): BOOLEAN;
VAR i: CARDINAL;
BEGIN
  IF pc + SIZE( x ) > len THEN RETURN FALSE END;
  i := 0;
  WHILE i < SIZE( x ) DO
    x[ i ] := prg[ pc ];
    INC( i ); INC( pc );
  END;
  RETURN TRUE;
END Fetch;

PROCEDURE Idx( i: TIndex ): TIndex;
BEGIN
  IF i >= vcMax THEN Err( 'Invalid index' ) END;
  RETURN i;
END Idx;

PROCEDURE Adr( a: TAddress ): TAddress;
BEGIN
  IF ( a > pcMax ) & ( a # aNone ) THEN Err( 'Invalid address' ) END;
  RETURN a;
END Adr;

PROCEDURE Push( x: TValue );
BEGIN
  IF ac >= acMax THEN Err( 'Stack overflow' ) END;
  INC( ac ); args[ ac ] := x;
END Push;

PROCEDURE Pop( VAR x: TValue );
BEGIN
  IF ac <= 0 THEN Err( 'Stack underflow' ) END;
  x := args[ ac ]; DEC( ac );
END Pop;

PROCEDURE Continue();
VAR x,y: TValue;
BEGIN
  WHILE pc < len DO
    pc0 := pc;
    IF ~ Fetch( cmd.fct ) THEN pc := pc0; RETURN END;
    CASE cmd.fct OF
    | fHlt: (* empty *)
            IF debug THEN Dbg() END;
            RETURN;
    | fVal: IF ~ Fetch( cmd.val ) THEN pc := pc0; RETURN END;
            IF debug THEN Dbg() END;
            Push( cmd.val );
    | fLod: IF ~ Fetch( cmd.idx ) THEN pc := pc0; RETURN END;
            IF debug THEN Dbg() END;
            Push( vars[Idx(cmd.idx)] );
    | fSto: IF ~ Fetch( cmd.idx ) THEN pc := pc0; RETURN END;
            IF debug THEN Dbg() END;
            Pop( vars[Idx(cmd.idx)] );
    | fOpr: IF ~ Fetch( cmd.opr ) THEN pc := pc0; RETURN END;
            IF debug THEN Dbg() END;
            CASE cmd.opr OF
            | oWrt: Pop( x ); FIO.WrLngInt( out, x, 0 ); FIO.WrLn( out );
            | oWrB: Pop( x ); FIO.WrStr( out, strs[x#vFalse] ); FIO.WrLn( out );
            | oNeg: Pop( x ); Push( -x );
            | oAdd: Pop( y ); Pop( x ); Push( x + y );
            | oSub: Pop( y ); Pop( x ); Push( x - y );
            | oMul: Pop( y ); Pop( x ); Push( x * y );
            | oDiv: Pop( y ); Pop( x );
                    IF y = 0 THEN Err( 'Division by zero' ) END;
                    Push( x DIV y );
            | oBus: Pop( y ); Pop( x ); Push( y - x );
            | oVid: Pop( y ); Pop( x );
                    IF x = 0 THEN Err( 'Division by zero' ) END;
                    Push( y DIV x );
            | oEql: Pop( y ); Pop( x ); Push( vals[x=y] );
            | oNeq: Pop( y ); Pop( x ); Push( vals[x#y] );
            | oLss: Pop( y ); Pop( x ); Push( vals[x<y] );
            | oGtr: Pop( y ); Pop( x ); Push( vals[x>y] );
            | oLeq: Pop( y ); Pop( x ); Push( vals[x<=y] );
            | oGeq: Pop( y ); Pop( x ); Push( vals[x>=y] );
            | oNot: Pop( x ); Push( TValue(TSet(x)/TSet(TValue(-1))) );
            | oOr : Pop( y ); Pop( x ); Push( TValue(TSet(x)+TSet(y)) );
            | oAnd: Pop( y ); Pop( x ); Push( TValue(TSet(x)*TSet(y)) );
            ELSE
              ErrCard( 'Unknown operation', ORD(cmd.opr) );
            END;
    | fJmp: IF ~ Fetch( cmd.adr ) OR ( cmd.adr = aNone ) THEN
              pc := pc0; RETURN
            END;
            IF debug THEN Dbg() END;
            pc := Adr( cmd.adr );
    | fJpn: IF ~ Fetch( cmd.adr ) OR ( cmd.adr = aNone ) THEN
              pc := pc0; RETURN
            END;
            IF debug THEN Dbg() END;
            Pop( x ); IF x # vFalse THEN pc := Adr( cmd.adr ) END;
    | fJpz: IF ~ Fetch( cmd.adr ) OR ( cmd.adr = aNone ) THEN
              pc := pc0; RETURN
            END;
            IF debug THEN Dbg() END;
            Pop( x ); IF x = vFalse THEN pc := Adr( cmd.adr ) END;
    ELSE
      ErrCard( 'Unknown function', ORD(cmd.fct) );
    END;
  END;
END Continue;

(***** programming *****)

PROCEDURE Write( x: ARRAY OF BYTE );
VAR i: CARDINAL;
BEGIN
  IF pos + SIZE( x ) > pcMax THEN Err( 'Program too large' ) END;
  FOR i := 0 TO HIGH( x ) DO prg[ pos ] := x[ i ]; INC( pos ) END;
  IF pos > len THEN len := pos END;
  IF immediateExec THEN Continue() END;
END Write;

PROCEDURE GetPos( VAR x: TAddress );
BEGIN
  x := pos;
END GetPos;

PROCEDURE SetPos( x: TAddress );
BEGIN
  IF x > len THEN Err( 'Invalid position' ) END;
  pos := x;
END SetPos;

PROCEDURE Read( VAR x: ARRAY OF BYTE );
VAR i: CARDINAL;
BEGIN
  IF pos + SIZE( x ) > len THEN Err( 'Nothing to read' ) END;
  FOR i := 0 TO HIGH( x ) DO x[ i ] := prg[ pos ]; INC( pos ) END;
END Read;

PROCEDURE Exec();
BEGIN
  pc := 0; Continue();
END Exec;

BEGIN
  immediateExec := FALSE;
  (* program *)
  Lib.Fill( ADR(prg), SIZE(prg), fHlt );
  pos := 0; len := 0;
  (* machine state *)
  pc := 0; pc0 := 0;
  Lib.Fill( ADR(args), SIZE(args), 0C );
  ac := 0;
  Lib.Fill( ADR(vars), SIZE(vars), 0C );
  (* module state *)
  first := TRUE;
  strs[ FALSE ] := 'FALSE';
  strs[ TRUE  ] := 'TRUE';
  vals[ FALSE ] := vFalse;
  vals[ TRUE  ] := vTrue;
END Comp.
